#define SPHO_ENABLE_DEBUG_PRINT

#include <sys/queue.h>

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "spho/spho.h"

#include "msph/tree.h"


struct tree_parse {
	struct msph_token_stream	*s;

	struct msph_token		curr;

	SLIST_HEAD(msph_token_l, msph_token)	head;
};

#define CURR_TOKEN(parse)	(&parse->curr)

#define CTX(parse)		((parse)->s->ctx)

static void	tree_parse_init(struct tree_parse *,
		    struct msph_token_stream *);
static int	tree_parse_next_token(struct tree_parse *);
static void	tree_parse_push_token(struct tree_parse *,
		    struct msph_token *);

static ssize_t	tree_ind_fprint(FILE *, int, struct msph_tree *);

static void	free_the_tree(struct msph_tree *);

static struct msph_tree_unit	*tree_parse_unit(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_body	*tree_parse_body(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_ident	*tree_parse_ident_extsize(struct tree_parse *,
				    struct msph_tree *, size_t);

static struct msph_tree_namedecl *tree_parse_namedecl(struct tree_parse *,
    struct msph_tree *);
static struct msph_tree_nameref *tree_parse_nameref(struct tree_parse *,
    struct msph_tree *);

static struct msph_tree_dir *tree_parse_tpdef(struct tree_parse *,
				struct msph_tree *);
static struct msph_tree_dir *tree_parse_assert(struct tree_parse *,
				  struct msph_tree *);
static struct msph_tree_dir *tree_parse_nomindecl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_dir *tree_parse_membdecl(
				    struct tree_parse *,
				    struct msph_tree *);


static struct msph_tree_tpexpr	*tree_parse_tpexpr(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_true(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_false(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_tpname_or_tpappl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_tpapplargs(struct tree_parse *,
				    struct msph_tree_nameref*,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_trait(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_conj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_disj(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_impl(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_arrow(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_box(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_forall(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_paren(struct tree_parse *,
				    struct msph_tree *);

struct msph_tree_root *
msph_tree_makeroot(struct msph_ctx *ctx)
{
	struct msph_tree_root	*root;

	if ((root = malloc(sizeof(*root))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(root)->type = MSPH_TREE_ROOT;
	TREE(root)->parent = NULL;

	root->ctx = ctx;

	STAILQ_INIT(&root->head);

	return (root);
}


/*
 * Expect to read next token, otherwise set error and do something
 * readres: (int) result of read
 * parse: struct tree_parse * to perform read
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_READ_NEXT(readres, parse, err_info, on_err) \
	do { \
		if (((readres) = tree_parse_next_token(parse)) == -1) \
			on_err; \
		if ((readres) == 0) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_EOF, \
			    err_info); \
			on_err; \
		} \
	} while (0)

/*
 * Expect current token type, otherwise set error and do something
 * parse: struct tree_parse * to perform read
 * toktype: token type expected
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_CURR_TOKEN(parse, toktype, err_info, on_err) \
	do { \
		if (CURR_TOKEN(parse)->type != toktype) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_NOMATCH, \
			    err_info); \
			on_err; \
		} \
	} while (0)

int
msph_tree_parse(struct msph_token_stream *s, struct msph_tree_root *root)
{
	struct tree_parse p;
	struct msph_tree_unit *unit;

	if (TREE(root)->type != MSPH_TREE_ROOT) {
		MSPH_ERR(s->ctx, MSPH_ERR_INVAL);
		return (-1);
	}

	tree_parse_init(&p, s);

	if ((unit = tree_parse_unit(&p, TREE(root))) == NULL)
		return (-1);

	STAILQ_INSERT_TAIL(&root->head, unit, entries);

	return (0);
}

ssize_t
msph_tree_fprint(FILE *f, struct msph_tree *tree) {
	return (tree_ind_fprint(f, 0, tree));
}

#define MSPH_TREE_INDENT	"    "

__attribute__((format(printf, 3, 4)))
static ssize_t
indent_fprintf(FILE *f, int indent, const char *fmt, ...)
{
	int res;
	ssize_t ret;
	int i;
	va_list ap;

	ret = 0;
	va_start(ap, fmt);

	for (i = 0; i < indent; i++) {
		res = fprintf(f, "%s", MSPH_TREE_INDENT);
		if (res < 0) {
			return ((ssize_t)res);
		}
		ret += (ssize_t)res;
	}

	res = vfprintf(f, fmt, ap);
	if (res < 0) {
		return ((ssize_t)res);
	}
	ret += (ssize_t)res;
	va_end(ap);

	return (ret);
}

static ssize_t
tree_ind_fprint(FILE *f, int indent, struct msph_tree *tree)
{
	ssize_t ret, res;
	struct msph_tree_root *root;
	struct msph_tree_unit *unit;
	struct msph_tree_body *body;
	struct msph_tree_dir *dir;
	struct msph_tree_nomindecl *nomd;
	struct msph_tree_assert *ass;
	struct msph_tree_tpdef *tpdef;
	struct msph_tree_membdecl *membdecl;
	struct msph_tree_conj *conj;
	struct msph_tree_disj *disj;
	struct msph_tree_impl *impl;
	struct msph_tree_arrow *arrow;
	struct msph_tree_box *box;
	struct msph_tree_forall *forall;
	struct msph_tree_tpappl *tpappl;
	struct msph_tree_trait *trait;
	struct msph_tree_tpname *name;
	struct msph_tree_paren *paren;
	struct msph_tree_ident *ident;
	struct msph_tree_tpexpr *tp;

	ret = 0;
	switch (TREE_ID(tree)) {
	case MSPH_TREE_ROOT:
		root = (struct msph_tree_root *)tree;
		if ((res = indent_fprintf(f, indent, "(root\n")) < 0)
			return (res);
		STAILQ_FOREACH(unit, &root->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(unit)))
			    < 0)
				return (res);
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		break;
	case MSPH_TREE_UNIT:
		unit = (struct msph_tree_unit *)tree;
		if ((res = indent_fprintf(f, indent, "(unit:%s\n", unit->name))
		    < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(unit->body))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BODY:
		body = (struct msph_tree_body *)tree;
		if ((res = indent_fprintf(f, indent, "(body\n")) < 0)
			return (res);
		ret += res;
		STAILQ_FOREACH(dir, &body->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(dir)))
			    < 0)
				return (res);
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NOMINDECL:
		nomd = (struct msph_tree_nomindecl *)tree;
		if ((res = indent_fprintf(f, indent, "(nomindecl:%u,%u\n",
		    TEXT(nomd)->pos.line, TEXT(nomd)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(nomd->name))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ASSERT:
		ass = (struct msph_tree_assert *)tree;
		if ((res = indent_fprintf(f, indent, "(assert:%u,%u\n",
		    TEXT(ass)->pos.line, TEXT(ass)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(ass->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(ass->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPDEF:
		tpdef = (struct msph_tree_tpdef *)tree;
		if ((res = indent_fprintf(f, indent, "(tpdef:%u,%u\n",
		    TEXT(tpdef)->pos.line, TEXT(tpdef)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpdef->name))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpdef->tp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_MEMBDECL:
		membdecl = (struct msph_tree_membdecl *)tree;
		if ((res = indent_fprintf(f, indent, "(membdecl:%u,%u\n",
		    TEXT(membdecl)->pos.line, TEXT(membdecl)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(membdecl->name)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(membdecl->tp)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_CONJ:
		conj = (struct msph_tree_conj *)tree;
		if ((res = indent_fprintf(f, indent, "(conj:%u,%u\n",
		    TEXT(conj)->pos.line, TEXT(conj)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(conj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(conj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_DISJ:
		disj = (struct msph_tree_disj *)tree;
		if ((res = indent_fprintf(f, indent, "(disj:%u,%u\n",
		    TEXT(disj)->pos.line, TEXT(disj)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(disj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(disj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_IMPL:
		impl = (struct msph_tree_impl *)tree;
		if ((res = indent_fprintf(f, indent, "(impl:%u,%u\n",
		    TEXT(impl)->pos.line, TEXT(impl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(impl->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(impl->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ARROW:
		arrow = (struct msph_tree_arrow *)tree;
		if ((res = indent_fprintf(f, indent, "(arrow:%u,%u\n",
		    TEXT(arrow)->pos.line, TEXT(arrow)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(arrow->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(arrow->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BOX:
		box = (struct msph_tree_box *)tree;
		if ((res = indent_fprintf(f, indent, "(box:%u,%u\n",
		    TEXT(box)->pos.line, TEXT(box)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(box->inner))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((res = indent_fprintf(f, indent, "(forall:%u,%u\n",
		    TEXT(forall)->pos.line, TEXT(forall)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(forall->name)))
		    < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(forall->inner)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_PAREN:
		paren = (struct msph_tree_paren *)tree;
		if ((res = indent_fprintf(f, indent, "(paren:%u,%u\n",
		    TEXT(paren)->pos.line, TEXT(paren)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(paren->inner)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPAPPL:
		tpappl = (struct msph_tree_tpappl *)tree;
		if ((res = indent_fprintf(f, indent, "(tpappl:%u,%u\n",
		    TEXT(tpappl)->pos.line, TEXT(tpappl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpappl->name))) < 0)
			return (res);
		ret += res;
		STAILQ_FOREACH(tp, &tpappl->args, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(tp))) < 0)
				return (res);
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPNAME:
		name = (struct msph_tree_tpname *)tree;
		if ((res = indent_fprintf(f, indent, "(name:%u,%u\n",
		    TEXT(name)->pos.line, TEXT(name)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(name->name))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;
		if ((res = indent_fprintf(f, indent, "(trait:%u,%u\n",
		    TEXT(trait)->pos.line, TEXT(trait)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(trait->body)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NAMEDECL:
	case MSPH_TREE_NAMEREF:
	case MSPH_TREE_IDENT:
		ident = IDENT(tree);
		if ((res = indent_fprintf(f, indent, "(ident:%u,%u:%s)\n",
		    TEXT(ident)->pos.line, TEXT(ident)->pos.col, ident->str))
		    < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TRUE:
		if ((res = indent_fprintf(f, indent, "(true:%u,%u)\n",
		    TEXT(tree)->pos.line, TEXT(tree)->pos.col)) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_FALSE:
		if ((res = indent_fprintf(f, indent, "(false:%u,%u)\n",
		    TEXT(tree)->pos.line, TEXT(tree)->pos.col)) < 0)
			return (res);
		ret += res;
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static void
tree_parse_init(struct tree_parse *p, struct msph_token_stream *s)
{
	p->s = s;

	memset(&p->curr, 0, sizeof(p->curr));

	SLIST_INIT(&p->head);
}

static int
tree_parse_next_token(struct tree_parse *p)
{
	int ret;
	struct msph_token *tok;
	if (! SLIST_EMPTY(&p->head)) {
		tok = SLIST_FIRST(&p->head);
		SLIST_REMOVE_HEAD(&p->head, entries);
		memcpy(&p->curr, tok, sizeof(p->curr));
		free(tok);
		ret = 1;
	} else {
		ret = (int)msph_token_stream_read(&p->curr, 1, p->s);
	}
#ifdef SPHO_ENABLE_DEBUG_PRINT
	if (ret == 1) {
		char buf[256];
		msph_token_str(buf, sizeof(buf), CURR_TOKEN(p));
		SPHO_DEBUG_PRINT("tree_parse_read_token: curr=%s\n",
		    buf);
	}
#endif
	return (ret);
}

struct msph_tree_unit *
tree_parse_unit(struct tree_parse *p, struct msph_tree *parent)
{
	size_t namelen;
	struct msph_tree_unit *unit;

	if ((unit = malloc(sizeof(*unit))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(unit)->type = MSPH_TREE_UNIT;
	TREE(unit)->parent = parent;
	DECOR_INIT(unit);
	unit->body = NULL;

	if ((namelen = strlcpy(unit->name, p->s->name, sizeof(unit->name)))
	    >= sizeof(unit->name)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TOOLONG);
		goto err;
	}

	if ((unit->body = tree_parse_body(p, TREE(unit))) == NULL)
		goto err;

	return (unit);

err:
	free_the_tree(TREE(unit));

	return (NULL);
}

static struct msph_tree_body *
tree_parse_body(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_dir *dir;
	struct msph_tree_body *body;
	struct msph_token *tok;

	if ((body = malloc(sizeof(*body))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(body)->type = MSPH_TREE_BODY;
	TREE(body)->parent = parent;
	STAILQ_INIT(&body->head);

	while ((res = tree_parse_next_token(p)) > 0) {
		dir = NULL;
		switch(CURR_TOKEN(p)->type) {
		case TOK_KW_TYPE:
			SPHO_DEBUG_PRINT("parsing tpdef\n");
			dir = tree_parse_tpdef(p, TREE(body));
			break;
		case TOK_KW_NOMINAL:
			dir = tree_parse_nomindecl(p, TREE(body));
			break;
		case TOK_KW_MEMBER:
			dir = tree_parse_membdecl(p, TREE(body));
			break;
		case TOK_KW_ASSERT:
			dir = tree_parse_assert(p, TREE(body));
			break;
		default:
			if ((tok = msph_token_copy(CTX(p), CURR_TOKEN(p)))
			    == NULL)
				goto err;
			tree_parse_push_token(p, tok);
			goto ret;
		}

		if (dir == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&body->head, dir, entries);
	}

	if (res == -1)
		goto err;
ret:
	return (body);
err:
	free_the_tree(TREE(body));

	return (NULL);
}

void
tree_parse_push_token(struct tree_parse *p, struct msph_token *tok)
{
	SLIST_INSERT_HEAD(&p->head, tok, entries);
	CURR_TOKEN(p)->type = TOK_INVAL;
}


struct msph_tree_dir *
tree_parse_tpdef(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpdef *def;
	struct msph_tree_namedecl *name;

	EXPECT_CURR_TOKEN(p, TOK_KW_TYPE, MSPH_TREE_TPDEF, return (NULL));

	if ((def = malloc(sizeof(*def))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(def)->type = MSPH_TREE_TPDEF;
	TREE(def)->parent = parent;
	TEXT(def)->pos = CURR_TOKEN(p)->pos;
	STAILQ_INIT(&def->params);
	DECOR_INIT(def);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->name = tree_parse_namedecl(p, TREE(def))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);

	if (CURR_TOKEN(p)->type == TOK_LBRAK) {
		ADD_FLAG(TREE(def), MSPH_TREE_FLAG_SCOPE);

		do {
			EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
			name = tree_parse_namedecl(p, TREE(def));
			if (name == NULL)
				goto err;
			STAILQ_INSERT_TAIL(&def->params, name, entries);

			EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
		} while (CURR_TOKEN(p)->type == TOK_COMMA);

		EXPECT_CURR_TOKEN(p, TOK_RBRAK, MSPH_TREE_TPDEF, goto err);
	}
	EXPECT_CURR_TOKEN(p, TOK_EQUALS, MSPH_TREE_TPDEF, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->tp = tree_parse_tpexpr(p, TREE(def))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)def);
err:
	free_the_tree(TREE(def));
	return (NULL);
}

struct msph_tree_namedecl *
tree_parse_namedecl(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_namedecl *name;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_NAMEDECL, return (NULL));

	name = (struct msph_tree_namedecl *)tree_parse_ident_extsize(p, parent,
	    sizeof(*name));
	if (name == NULL)
		return (NULL);

	SPHO_ASSERT(TREE(name)->type == MSPH_TREE_IDENT);
	TREE(name)->type |= MSPH_TREE_NAMEDECL;

	return (name);
}

struct msph_tree_nameref *
tree_parse_nameref(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_nameref *name;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_NAMEREF, return (NULL));

	name = (struct msph_tree_nameref *)tree_parse_ident_extsize(p, parent,
	    sizeof(*name));
	if (name == NULL)
		return (NULL);

	SPHO_ASSERT(TREE(name)->type == MSPH_TREE_IDENT);
	TREE(name)->type |= MSPH_TREE_NAMEREF;

	return (name);
}

struct msph_tree_ident *
tree_parse_ident_extsize(struct tree_parse *p, struct msph_tree *parent,
    size_t size)
{
	struct msph_tree_ident *id;
	SPHO_PRECOND(size >= sizeof(struct msph_tree_ident));

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_IDENT, return (NULL));

	if ((id = malloc(sizeof(size))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(id)->type = MSPH_TREE_IDENT;
	TREE(id)->parent = parent;
	TEXT(id)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(id);

	if (SPHO_STRLCPY(id->str, CURR_TOKEN(p)->data.str, sizeof(id->str))
	    >= sizeof(id->str)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TREE_TOOLONG);
		goto err;
	}

	return (id);
err:
	free_the_tree(TREE(id));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_tpexpr(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_tpexpr *tp;

	switch (CURR_TOKEN(p)->type) {
	case TOK_AMP:
		tp = tree_parse_conj(p, parent);
		break;
	case TOK_PIPE:
		tp = tree_parse_disj(p, parent);
		break;
	case TOK_IMPL:
		tp = tree_parse_impl(p, parent);
		break;
	case TOK_RARROW:
		tp = tree_parse_arrow(p, parent);
		break;
	case TOK_KW_BOX:
		tp = tree_parse_box(p, parent);
		break;
	case TOK_KW_FORALL:
		tp = tree_parse_forall(p, parent);
		break;
	case TOK_IDENT:
		tp = tree_parse_tpname_or_tpappl(p, parent);
		break;
	case TOK_CONST_TRUE:
		tp = tree_parse_true(p, parent);
		break;
	case TOK_CONST_FALSE:
		tp = tree_parse_false(p, parent);
		break;
	case TOK_LPAREN:
		tp = tree_parse_paren(p, parent);
		break;
	case TOK_LBRACE:
		tp = tree_parse_trait(p, parent);
		break;
	default:
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH, MSPH_TREE_TPEXPR);
		tp = NULL;
	}

	SPHO_DEBUG_PRINT("returning tp type=%x\n", TREE(tp)->type);
	return (tp);
}

struct msph_tree_tpexpr *
tree_parse_conj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_conj *conj;

	EXPECT_CURR_TOKEN(p, TOK_AMP, MSPH_TREE_CONJ, return (NULL));

	if ((conj = malloc(sizeof(*conj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(conj)->type = MSPH_TREE_CONJ;
	TREE(conj)->parent = parent;
	TEXT(conj)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(conj);
	conj->ltp = NULL;
	conj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->ltp = tree_parse_tpexpr(p, TREE(conj))) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->rtp = tree_parse_tpexpr(p, TREE(conj)))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)conj);
err:
	free_the_tree((struct msph_tree *)conj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_disj(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_disj *disj;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_DISJ, return (NULL));

	if ((disj = malloc(sizeof(*disj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(disj)->type = MSPH_TREE_DISJ;
	TREE(disj)->parent = parent;
	TEXT(disj)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(disj);
	disj->ltp = NULL;
	disj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->ltp = tree_parse_tpexpr(p, (struct msph_tree *)disj)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->rtp = tree_parse_tpexpr(p, (struct msph_tree *)disj))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)disj);
err:
	free_the_tree((struct msph_tree *)disj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_impl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_impl *impl;

	EXPECT_CURR_TOKEN(p, TOK_IMPL, MSPH_TREE_IMPL, return (NULL));

	if ((impl = malloc(sizeof(*impl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(impl)->type = MSPH_TREE_IMPL;
	TREE(impl)->parent = parent;
	TEXT(impl)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(impl);
	impl->ltp = NULL;
	impl->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->ltp = tree_parse_tpexpr(p, (struct msph_tree *)impl)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->rtp = tree_parse_tpexpr(p, (struct msph_tree *)impl))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)impl);
err:
	free_the_tree((struct msph_tree *)impl);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_arrow(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_arrow *arrow;

	EXPECT_CURR_TOKEN(p, TOK_RARROW, MSPH_TREE_ARROW, return (NULL));

	if ((arrow = malloc(sizeof(*arrow))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(arrow)->type = MSPH_TREE_ARROW;
	TREE(arrow)->parent = parent;
	TEXT(arrow)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(arrow);
	arrow->ltp = NULL;
	arrow->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->ltp = tree_parse_tpexpr(p, (struct msph_tree *)arrow)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->rtp = tree_parse_tpexpr(p, (struct msph_tree *)arrow))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)arrow);
err:
	free_the_tree((struct msph_tree *)arrow);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_box(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_box *box;

	EXPECT_CURR_TOKEN(p, TOK_KW_BOX, MSPH_TREE_BOX, return (NULL));

	if ((box = malloc(sizeof(*box))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(box)->type = MSPH_TREE_BOX;
	TREE(box)->parent = parent;
	TEXT(box)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(box);
	box->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_BOX, goto err);
	if ((box->inner = tree_parse_tpexpr(p, TREE(box))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)box);
err:
	free_the_tree(TREE(box));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_forall(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_forall *fa;

	EXPECT_CURR_TOKEN(p, TOK_KW_FORALL, MSPH_TREE_FORALL, return (NULL));

	if ((fa = malloc(sizeof(*fa))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(fa)->type = MSPH_TREE_FORALL;
	TREE(fa)->parent = parent;
	TEXT(fa)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(fa);
	fa->name = NULL;
	fa->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->name = tree_parse_namedecl(p, TREE(fa))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_DOT, MSPH_TREE_FORALL, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->inner = tree_parse_tpexpr(p, TREE(fa))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)fa);
err:
	free_the_tree(TREE(fa));
	return (NULL);
}

/* parse name or appl */
struct msph_tree_tpexpr *
tree_parse_tpname_or_tpappl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpname *tpname;
	struct msph_tree_tpexpr *ret;
	struct msph_tree_nameref *nameref;
	struct msph_token *tok;

	ret = NULL;
	nameref = NULL;

	if ((nameref = tree_parse_nameref(p, NULL)) == NULL)
		return (NULL);

	if ((res = tree_parse_next_token(p)) == 1) {
		if (CURR_TOKEN(p)->type == TOK_LBRAK) {
			if ((ret = tree_parse_tpapplargs(p, nameref, parent))
			    == NULL)
				goto err;
			return (ret);
		}

		if ((tok = msph_token_copy(CTX(p), CURR_TOKEN(p)))
		    == NULL) {
			goto err;
		}
		tree_parse_push_token(p, tok);
	} else if (res == -1) {
		goto err;
	}

	if ((tpname = malloc(sizeof(*tpname))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}
	TREE(tpname)->type = MSPH_TREE_TPNAME;
	TREE(tpname)->parent = parent;

	TEXT(tpname)->pos = TEXT(nameref)->pos;
	DECOR_INIT(tpname);
	TREE(nameref)->parent = TREE(tpname);
	tpname->name = nameref;

	return ((struct msph_tree_tpexpr *)tpname);
err:
	free_the_tree(TREE(nameref));

	return (NULL);
}

/**
 * Parse argument list part [...], of type application A[...].
 *
 * @param p Parse state
 * @param id Parsed ident A (on failure, this is not freed)
 * @param parent Parent of resulting appl node
 *
 * @return tpexpr node representing A[...].
 */
struct msph_tree_tpexpr *
tree_parse_tpapplargs(struct tree_parse *p, struct msph_tree_nameref *name,
    struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpappl *tpappl;
	struct msph_tree_tpexpr *tp;

	tp = NULL;

	EXPECT_CURR_TOKEN(p, TOK_LBRAK, MSPH_TREE_TPAPPL, return (NULL));

	if ((tpappl = malloc(sizeof(*tpappl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(tpappl)->type = MSPH_TREE_TPAPPL;
	TREE(tpappl)->parent = parent;
	DECOR_INIT(tpappl);
	tpappl->name = NULL;
	STAILQ_INIT(&tpappl->args);

	/* parse the argument list */
	while ((res = tree_parse_next_token(p)) == 1) {
		/* if ']' we have reached end of arguments */
		if (CURR_TOKEN(p)->type == TOK_RBRAK)
			break;

		/* if not at start of argument list, check for comma to know if
		 * we should parse another argument */
		if (tp != NULL) {
			EXPECT_CURR_TOKEN(p, TOK_COMMA, MSPH_TREE_TPAPPL,
			    goto err);
			EXPECT_READ_NEXT(res, p, MSPH_TREE_TPAPPL, goto err);
		}

		if ((tp = tree_parse_tpexpr(p, TREE(tpappl))) == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&tpappl->args, tp, entries);
	}

	/* SYS_ERROR set in tree_parse_next_token */
	if (res == -1)
		goto err;

	/* did not read final ']' because of EOF */
	if (res == 0) {
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_EOF, MSPH_TREE_TPAPPL);
		goto err;
	}

	/* finally, appl takes ownership of id */
	TREE(name)->parent = TREE(tpappl);
	tpappl->name = name;
	TEXT(tpappl)->pos = TEXT(name)->pos;

	return ((struct msph_tree_tpexpr *)tpappl);
err:
	free_the_tree(TREE(tpappl));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_true(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *t;

	EXPECT_CURR_TOKEN(p, TOK_CONST_TRUE, MSPH_TREE_TRUE, return (NULL));

	if ((t = malloc(sizeof(*t))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(t)->type = MSPH_TREE_TRUE;
	TREE(t)->parent = parent;
	TEXT(t)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(t);

	return ((struct msph_tree_tpexpr *)t);
}

struct msph_tree_tpexpr *
tree_parse_false(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *f;

	EXPECT_CURR_TOKEN(p, TOK_CONST_FALSE, MSPH_TREE_FALSE, return (NULL));

	if ((f = malloc(sizeof(*f))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(f)->type = MSPH_TREE_FALSE;
	TREE(f)->parent = parent;
	TEXT(f)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(f);

	return ((struct msph_tree_tpexpr *)f);
}

struct msph_tree_tpexpr *
tree_parse_paren(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_paren *par;

	EXPECT_CURR_TOKEN(p, TOK_LPAREN, MSPH_TREE_FALSE, return (NULL));

	if ((par = malloc(sizeof(*par))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(par)->type = MSPH_TREE_PAREN;
	TREE(par)->parent = parent;
	TEXT(par)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(par);
	par->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);

	if ((par->inner = tree_parse_tpexpr(p, TREE(par))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RPAREN, MSPH_TREE_PAREN, goto err);

	return ((struct msph_tree_tpexpr *)par);
err:
	free_the_tree(TREE(par));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_trait(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_trait *trait;

	EXPECT_CURR_TOKEN(p, TOK_LBRACE, MSPH_TREE_TRAIT, return (NULL));

	if ((trait = malloc(sizeof(*trait))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(trait)->type = MSPH_TREE_TRAIT;
	TREE(trait)->parent = parent;
	TEXT(trait)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(trait);
	trait->body = NULL;

	if ((trait->body = tree_parse_body(p, TREE(trait))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TRAIT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RBRACE, MSPH_TREE_TRAIT, goto err);

	return ((struct msph_tree_tpexpr *)trait);
err:
	free_the_tree(TREE(trait));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_nomindecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_namedecl *name;
	struct msph_tree_nomindecl *nomind;

	EXPECT_CURR_TOKEN(p, TOK_KW_NOMINAL, MSPH_TREE_NOMINDECL,
	    return (NULL));

	if ((nomind = malloc(sizeof(*nomind))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(nomind)->type = MSPH_TREE_NOMINDECL;
	TREE(nomind)->parent = parent;
	TEXT(nomind)->pos = CURR_TOKEN(p)->pos;
	nomind->name = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_NOMINDECL, goto err);
	if ((name = tree_parse_namedecl(p, TREE(nomind))) == NULL)
		goto err;

	nomind->name = name;

	return ((struct msph_tree_dir *)nomind);
err:
	free_the_tree(TREE(nomind));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_membdecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_membdecl *membd;

	EXPECT_CURR_TOKEN(p, TOK_KW_MEMBER, MSPH_TREE_MEMBDECL,
	    return (NULL));

	if ((membd = malloc(sizeof(*membd))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(membd)->type = MSPH_TREE_MEMBDECL;
	TREE(membd)->parent = parent;
	TEXT(membd)->pos = CURR_TOKEN(p)->pos;
	membd->name = NULL;
	membd->tp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	if ((membd->name = tree_parse_namedecl(p, TREE(membd))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_COLON, MSPH_TREE_MEMBDECL,
	    goto err);
	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);

	if ((membd->tp = tree_parse_tpexpr(p, TREE(membd))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)membd);
err:
	free_the_tree(TREE(membd));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_assert(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_assert *ass;

	EXPECT_CURR_TOKEN(p, TOK_KW_ASSERT, MSPH_TREE_ASSERT,
	    return (NULL));

	if ((ass = malloc(sizeof(*ass))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(ass)->type = MSPH_TREE_ASSERT;
	TREE(ass)->parent = parent;
	TEXT(ass)->pos = CURR_TOKEN(p)->pos;
	ass->ltp = NULL;
	ass->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);
	if ((ass->ltp = tree_parse_tpexpr(p, TREE(ass))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_SUB, MSPH_TREE_ASSERT,
	    goto err);
	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);

	if ((ass->rtp = tree_parse_tpexpr(p, TREE(ass))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)ass);
err:
	free_the_tree(TREE(ass));

	return (NULL);
}

static void
free_the_tree(struct msph_tree *tree)
{
	/* will anyone ever free the tree? :´( */
	SPHO_POSTCOND((void *)tree != (void *)free);
	// SPHO_POSTCOND(0);
}

